import re
import hashlib
import json
import singer

LOGGER = singer.get_logger()

# Convert camelCase to snake_case
def convert(name):
    regsub = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', regsub).lower()


# Convert keys in json array
def convert_array(arr):
    new_arr = []
    for i in arr:
        if isinstance(i, list):
            new_arr.append(convert_array(i))
        elif isinstance(i, dict):
            new_arr.append(convert_json(i))
        else:
            new_arr.append(i)
    return new_arr


# Convert keys in json
def convert_json(this_json):
    out = {}
    for key in this_json:
        new_key = convert(key)
        if isinstance(this_json[key], dict):
            out[new_key] = convert_json(this_json[key])
        elif isinstance(this_json[key], list):
            out[new_key] = convert_array(this_json[key])
        else:
            out[new_key] = this_json[key]
    return out


# Remove 'keys' node, if exists
def remove_keys_nodes(this_json, path):
    new_json = this_json
    i = 0
    for record in list(this_json[path]):
        if record.get('keys', None):
            new_json[path][i].pop('keys')
        i = i + 1
    return new_json


# Create MD5 hash key for data element
def hash_data(data):
    # Prepare the project id hash
    hashId = hashlib.md5()
    hashId.update(repr(data).encode('utf-8'))
    return hashId.hexdigest()


# Denest keys values list to dimension_list keys
def denest_key_fields(this_json, stream_name, path, dimensions_list):
    new_json = this_json
    i = 0
    for record in list(this_json[path]):
        for key in list(record.keys()):
            if isinstance(record[key], list):
                if key == 'keys':
                    dim_num = 0
                    # Add dimensions_hash_key for performance_report_custom
                    if stream_name == 'performance_report_custom':
                        dims_md5 = str(hash_data(json.dumps(record[key], sort_keys=True)))
                        new_json[path][i]['dimensions_hash_key'] = dims_md5
                    for dimension in dimensions_list:
                        new_json[path][i][dimension] = record[key][dim_num]
                        dim_num = dim_num + 1
        i = i + 1
    return new_json


# Add site_url to results
def add_site_url(this_json, path, site):
    new_json = this_json
    i = 0
    for record in this_json[path]:
        new_json[path][i]['site_url'] = site
        i = i + 1
    return new_json


# Add search_type to results
def add_search_type(this_json, path, sub_type):
    new_json = this_json
    i = 0
    for record in this_json[path]:
        new_json[path][i]['search_type'] = sub_type
        i = i + 1
    return new_json


def transform_reports(this_json, stream_name, path, site, sub_type, dimensions_list):
    # de-nest keys array to dimension fields and add MD5 hash key for custom report
    denested_json = denest_key_fields(this_json, stream_name, path, dimensions_list)
    # remove keys array node
    keyless_json = remove_keys_nodes(denested_json, path)
    # add site_url and search_type to results
    new_json = add_search_type(add_site_url(keyless_json, path, site), path, sub_type)
    return new_json


# Run all transforms: convert camelCase to snake_case for fieldname keys,
#  and stream-specific transforms for sitemaps and performance_reports.
def transform_json(this_json, stream_name, path, site, sub_type, dimensions_list):
    converted_json = convert_json(this_json)
    if stream_name == 'sitemaps':
        new_json = add_site_url(converted_json, path, site)
    elif stream_name.startswith('performance_report'):
        new_json = transform_reports(converted_json, stream_name, path, site, sub_type, dimensions_list)
    else:
        new_json = converted_json
    return new_json



# streams: API URL endpoints to be called
# properties:
#   <root node>: Plural stream name for the endpoint
#   path: API endpoint relative path, when added to the base URL, creates the full path
#   key_properties: Primary key fields for identifying an endpoint record.
#   replication_method: INCREMENTAL or FULL_TABLE
#   replication_keys: bookmark_field(s), typically a date-time, used for filtering the results
#        and setting the state
#   data_key: JSON element containing the records for the endpoint
#   api_method: GET or POST; default = 'GET'
#   params: Query, sort, and other endpoint specific parameters; default = {}
#   pagination: types are none, body, params; default = 'none'
#       none = no pagination
#       body = POST has startRow and rowLimit in body payload
#       params = GET has startRow and rowLimit in URL query params
#   sub_types: list of sub_types for endpoint looping; delfault = ['self']
#   bookmark_type: Data type for bookmark, integer or datetime

STREAMS = {
    'sites': {
        'key_properties': ['site_url'],
        'replication_method': 'FULL_TABLE',
        'path': 'sites/{}',
        'data_key': 'site_entry',
        'api_method': 'GET'
    },

    'sitemaps': {
        'key_properties': ['site_url', 'path', 'last_submitted'],
        'replication_method': 'FULL_TABLE',
        'path': 'sites/{}/sitemaps',
        'data_key': 'sitemap',
        'api_method': 'GET'
    },

    'performance_report_custom': {
        'key_properties': ['site_url', 'search_type', 'date', 'dimensions_hash_key'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'auto'
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },

    'performance_report_date': {
        'key_properties': ['site_url', 'search_type', 'date'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'byProperty',
            'dimensions': ['date']
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },

    'performance_report_country': {
        'key_properties': ['site_url', 'search_type', 'date', 'country'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'byProperty',
            'dimensions': ['date', 'country']
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },

    'performance_report_device': {
        'key_properties': ['site_url', 'search_type', 'date', 'device'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'byProperty',
            'dimensions': ['date', 'device']
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },

    'performance_report_page': {
        'key_properties': ['site_url', 'search_type', 'date', 'page'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date', 'page'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'byPage',
            'dimensions': ['date', 'page']
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },

    'performance_report_query': {
        'key_properties': ['site_url', 'search_type', 'date', 'query'],
        'replication_method': 'INCREMENTAL',
        'replication_keys': ['date'],
        'path': 'sites/{}/searchAnalytics/query',
        'data_key': 'rows',
        'api_method': 'POST',
        'row_limit': 10000,
        'body': {
            'aggregationType': 'byProperty',
            'dimensions': ['date', 'query']
        },
        'bookmark_type': 'datetime',
        'pagination': 'body',
        'sub_types': ['web', 'image', 'video']
    },
}



import json
from datetime import datetime, timedelta
from urllib.parse import quote
import backoff
import requests
from requests.exceptions import Timeout, ConnectionError

import singer
from singer import metrics
from singer import utils

BASE_URL = 'https://www.googleapis.com/webmasters/v3'
GOOGLE_TOKEN_URI = 'https://oauth2.googleapis.com/token'
LOGGER = singer.get_logger()

# set default timeout of 300 seconds
REQUEST_TIMEOUT = 300

class GoogleError(Exception):
    pass

class Server5xxError(GoogleError):
    pass

class GoogleBadRequestError(GoogleError):
    pass

class GoogleUnauthorizedError(GoogleError):
    pass

class GooglePaymentRequiredError(GoogleError):
    pass

class GoogleNotFoundError(GoogleError):
    pass

class GoogleMethodNotAllowedError(GoogleError):
    pass

class GoogleConflictError(GoogleError):
    pass

class GoogleGoneError(GoogleError):
    pass

class GooglePreconditionFailedError(GoogleError):
    pass

class GoogleRequestEntityTooLargeError(GoogleError):
    pass

class GoogleRequestedRangeNotSatisfiableError(GoogleError):
    pass

class GoogleExpectationFailedError(GoogleError):
    pass

class GoogleForbiddenError(GoogleError):
    pass

class GoogleUnprocessableEntityError(GoogleError):
    pass

class GooglePreconditionRequiredError(GoogleError):
    pass

class GoogleRateLimitExceeded(GoogleError):
    pass

class GoogleInternalServiceError(Server5xxError):
    pass

class GoogleNotImplementedError(Server5xxError):
    pass

class GoogleServiceUnavailable(Server5xxError):
    pass

class GoogleQuotaExceededError(GoogleError):
    pass

# Error Codes: https://developers.google.com/webmaster-tools/search-console-api-original/v3/errors
ERROR_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": GoogleBadRequestError,
        "message": "The request is missing or has bad parameters."
    },
    401: {
        "raise_exception": GoogleUnauthorizedError,
        "message": "Invalid authorization credentials."
    },
    402: {
        "raise_exception": GooglePaymentRequiredError,
        "message": "The requested operation requires more resources than the quota allows. Payment is required to complete the operation."
    },
    403: {
        "raise_exception": GoogleForbiddenError,
        "message": "Invalid authorization credentials or permissions."
    },
    404: {
        "raise_exception": GoogleNotFoundError,
        "message": "The requested resource does not exist."
    },
    405: {
        "raise_exception": GoogleMethodNotAllowedError,
        "message": "The HTTP method associated with the request is not supported."
    },
    409: {
        "raise_exception": GoogleConflictError,
        "message": "The API request cannot be completed because the requested operation would conflict with an existing item."
    },
    410: {
        "raise_exception": GoogleGoneError,
        "message": "The requested resource is permanently unavailable."
    },
    412: {
        "raise_exception": GooglePreconditionFailedError,
        "message": "The condition set in the request's If-Match or If-None-Match HTTP request header was not met."
    },
    413: {
        "raise_exception": GoogleRequestEntityTooLargeError,
        "message": "The request is too large."
    },
    416: {
        "raise_exception": GoogleRequestedRangeNotSatisfiableError,
        "message": "The request specified a range that cannot be satisfied."
    },
    417: {
        "raise_exception": GoogleExpectationFailedError,
        "message": "A client expectation cannot be met by the server."
    },
    422: {
        "raise_exception": GoogleUnprocessableEntityError,
        "message": "The request was not able to process right now."
    },
    428: {
        "raise_exception": GooglePreconditionRequiredError,
        "message": "The request requires a precondition If-Match or If-None-Match which is not provided."
    },
    429: {
        "raise_exception": GoogleRateLimitExceeded,
        "message": "Rate limit has been exceeded."
    },
    500: {
        "raise_exception": GoogleInternalServiceError,
        "message": "The request failed due to an internal error."
    },
    501: {
        "raise_exception": GoogleNotImplementedError,
        "message": "Functionality does not exist."
    },
    503: {
        "raise_exception": GoogleServiceUnavailable,
        "message": "The API service is currently unavailable."
    }
}

def raise_for_error(response):
    # Forming a response message for raising custom exception

    try:
        response_json = response.json()
    except Exception:
        response_json = {}

    error_code = response.status_code
    error_message = response_json.get("error_description") or response_json.get("error",
        ERROR_CODE_EXCEPTION_MAPPING.get(error_code,
            {})).get("message", "An Unknown Error occurred, please try after some time.")

    message = "HTTP-error-code: {}, Error: {}".format(
        error_code, error_message)

    # Raise GoogleQuotaExceededError if 403 error code returned due to QuotaExceeded
    response_error = json.dumps(response_json.get('error', error_message))
    if error_code == 403 and "quotaExceeded" in response_error:
        ex = GoogleQuotaExceededError
    else:
        ex = ERROR_CODE_EXCEPTION_MAPPING.get(error_code, {}).get("raise_exception", GoogleError)
    raise ex(message) from None


class GoogleClient: # pylint: disable=too-many-instance-attributes
    def __init__(self,
                 client_id,
                 client_secret,
                 refresh_token,
                 site_urls,
                 user_agent=None,
                 timeout_from_config=REQUEST_TIMEOUT):
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__refresh_token = refresh_token
        self.__site_urls = site_urls
        self.__user_agent = user_agent
        self.__access_token = None
        self.__expires = None
        self.__session = requests.Session()
        self.base_url = None

        # if the 'timeout_from_config' value is 0, "0", "" or not passed then set default value of 300 seconds.
        if timeout_from_config and float(timeout_from_config):
            # update the request timeout for the requests
            self.request_timeout = float(timeout_from_config)
        else:
            # set the default timeout of 300 seconds
            self.request_timeout = REQUEST_TIMEOUT

    def check_sites_access(self):
        site_list = self.__site_urls.replace(" ", "").split(",")

        for site in site_list:
            site_encoded = quote(site, safe='')

            #Sample query API for check site access
            path = 'sites/{}/searchAnalytics/query'.format(site_encoded)
            body = {'startDate': '2021-04-01', 'endDate': '2021-05-01'}
            self.post(path=path, data=json.dumps(body))

    # during 'Timeout' error there is also possibility of 'ConnectionError',
    # hence added backoff for 'ConnectionError' too.
    @backoff.on_exception(backoff.expo,
                          (Timeout, ConnectionError),
                          max_tries=5,
                          factor=2)
    def __enter__(self):
        self.get_access_token()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    @backoff.on_exception(backoff.expo,
                          Server5xxError,
                          max_tries=5,
                          factor=2)
    def get_access_token(self):
        # The refresh_token never expires and may be used many times to generate each access_token
        # Since the refresh_token does not expire, it is not included in get access_token response
        if self.__access_token is not None and self.__expires > datetime.utcnow():
            return

        headers = {}
        if self.__user_agent:
            headers['User-Agent'] = self.__user_agent

        response = self.__session.post(
            url=GOOGLE_TOKEN_URI,
            headers=headers,
            data={
                'grant_type': 'refresh_token',
                'client_id': self.__client_id,
                'client_secret': self.__client_secret,
                'refresh_token': self.__refresh_token,
            },
            timeout=self.request_timeout)

        if response.status_code != 200:
            raise_for_error(response)

        data = response.json()
        self.__access_token = data['access_token']
        self.__expires = datetime.utcnow() + timedelta(seconds=data['expires_in'])
        LOGGER.info('Authorized, token expires = {}'.format(self.__expires))

    @backoff.on_exception(backoff.constant,
                          GoogleQuotaExceededError,
                          max_tries=2,  # Only retry once
                          interval=900,  # Backoff for 15 minutes in case of Quota Exceeded error
                          jitter=None)  # Interval value not consistent if jitter not None
    # backoff for 5 times, with 10 seconds consistent interval
    @backoff.on_exception(backoff.constant,
                          Timeout,
                          max_tries=5,
                          interval=10,
                          jitter=None) # Interval value not consistent if jitter not None
    @backoff.on_exception(backoff.expo,
                          (Server5xxError, ConnectionError, GoogleRateLimitExceeded),
                          max_tries=7,
                          factor=3)
    # Rate Limit:
    #  https://developers.google.com/webmaster-tools/search-console-api-original/v3/limits
    @utils.ratelimit(1200, 60)
    def request(self, method, path=None, url=None, **kwargs):

        self.get_access_token()

        if not url and self.base_url is None:
            self.base_url = BASE_URL

        if not url and path:
            url = '{}/{}'.format(self.base_url, path)

        # endpoint = stream_name (from sync.py API call)
        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
        else:
            endpoint = None

        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        kwargs['headers']['Authorization'] = 'Bearer {}'.format(self.__access_token)

        if self.__user_agent:
            kwargs['headers']['User-Agent'] = self.__user_agent

        if method == 'POST':
            kwargs['headers']['Content-Type'] = 'application/json'

        with metrics.http_request_timer(endpoint) as timer:
            response = self.__session.request(method, url, timeout=self.request_timeout, **kwargs)
            timer.tags[metrics.Tag.http_status_code] = response.status_code

        if response.status_code != 200:
            raise_for_error(response)

        return response.json()

    def get(self, path, **kwargs):
        return self.request('GET', path=path, **kwargs)

    def post(self, path, **kwargs):
        return self.request('POST', path=path, **kwargs)



import json
from datetime import datetime, timedelta
from urllib.parse import quote
import singer
from singer import metrics, metadata, Transformer, utils
from singer.utils import strptime_to_utc, strftime
from tap_google_search_console.transform import transform_json
from tap_google_search_console.streams import STREAMS

LOGGER = singer.get_logger()
BASE_URL = 'https://www.googleapis.com/webmasters/v3'
# Google Search Console is generally delayed 2-3 days
# However, delays up to 10 days have occurred in the past 6 months (late 2019, early 2020)
# Reference: https://support.google.com/webmasters/answer/96568?hl=en
ATTRIBUTION_DAYS = 14
DATE_WINDOW_SIZE = 30

def write_schema(catalog, stream_name):
    stream = catalog.get_stream(stream_name)
    schema = stream.schema.to_dict()
    try:
        singer.write_schema(stream_name, schema, stream.key_properties)
    except OSError as err:
        LOGGER.error('OS Error writing schema for: {}'.format(stream_name))
        LOGGER.error('Error: {}'.format(err))
        raise err


def write_record(stream_name, record, time_extracted):
    try:
        singer.messages.write_record(stream_name, record, time_extracted=time_extracted)
    except OSError as err:
        LOGGER.error('OS Error writing record for: {}'.format(stream_name))
        LOGGER.error('Error: {}'.format(err))
        raise err


def get_bookmark(state, stream, site, sub_type, default):
    if (state is None) or ('bookmarks' not in state):
        return default
    return (
        state
        .get('bookmarks', {})
        .get(stream, {})
        .get(site, {})
        .get(sub_type, default)
    )


def write_bookmark(state, stream, site, sub_type, value):
    if 'bookmarks' not in state:
        state['bookmarks'] = {}
    if stream not in state['bookmarks']:
        state['bookmarks'][stream] = {}
    if site not in state['bookmarks'][stream]:
        state['bookmarks'][stream][site] = {}
    state['bookmarks'][stream][site][sub_type] = value
    LOGGER.info('Write state for Stream: {}, Site: {}, Type: {}, value: {}'.format(
        stream, site, sub_type, value))
    singer.write_state(state)


def process_records(catalog, #pylint: disable=too-many-branches
                    stream_name,
                    records,
                    time_extracted,
                    bookmark_field=None,
                    max_bookmark_value=None,
                    last_datetime=None):
    stream = catalog.get_stream(stream_name)
    schema = stream.schema.to_dict()
    stream_metadata = metadata.to_map(stream.metadata)

    with metrics.record_counter(stream_name) as counter:
        for record in records:
            # Transform record for Singer.io
            with Transformer() as transformer:
                transformed_record = transformer.transform(
                    record,
                    schema,
                    stream_metadata)

                # Reset max_bookmark_value to new value if higher
                if bookmark_field and (bookmark_field in transformed_record):
                    bookmark_date = transformed_record.get(bookmark_field)
                    bookmark_dttm = strptime_to_utc(bookmark_date)
                    last_dttm = strptime_to_utc(last_datetime)

                    if not max_bookmark_value:
                        max_bookmark_value = last_datetime

                    max_bookmark_dttm = strptime_to_utc(max_bookmark_value)

                    if bookmark_dttm > max_bookmark_dttm:
                        max_bookmark_value = strftime(bookmark_dttm)

                    # Keep only records whose bookmark is after the last_datetime
                    if bookmark_dttm >= last_dttm:
                        write_record(stream_name, transformed_record, \
                            time_extracted=time_extracted)
                        counter.increment()
                else:
                    write_record(stream_name, transformed_record, time_extracted=time_extracted)
                    counter.increment()

        LOGGER.info('Stream: {}, Processed {} records'.format(stream_name, counter.value))
        return max_bookmark_value


# Sync a specific endpoint
# For performance reports, this is for: reprot, site, sub_type, date window
def sync_endpoint(client, #pylint: disable=too-many-branches
                  catalog,
                  state,
                  start_date,
                  stream_name,
                  site,
                  sub_type,
                  dimensions_list,
                  path,
                  endpoint_config,
                  api_method,
                  pagination,
                  static_params,
                  bookmark_field=None,
                  data_key=None,
                  body_params=None,
                  id_fields=None):

    # Get the latest bookmark for the stream and set the last_datetime
    last_datetime = None
    max_bookmark_value = None

    last_datetime = get_bookmark(state, stream_name, site, sub_type, start_date)
    max_bookmark_value = last_datetime

    # Pagination: loop thru all pages of data
    # Pagination types: none, body, params
    # Each page has an offset (starting value) and a limit (batch size, number of records)
    # Increase the "offset" by the "limit" for each batch.
    # Continue until the "offset" exceeds the total_records.
    offset = 0 # Starting offset value for each batch API call
    limit = endpoint_config.get('row_limit', 1000) # Batch size; Number of records per API call
    total_records = 0
    batch_count = limit
    page = 1

    while limit == batch_count:
        if pagination == 'body':
            body = {
                'startRow': offset,
                'rowLimit': limit,
                **body_params # adds in endpoint specific, sort, filter body params
            }
            params = static_params
        elif pagination == 'params':
            params = {
                'startRow': offset,
                'rowLimit': limit,
                **static_params # adds in endpoint specific, sort, filter body params
            }
            body = body_params
        else:
            params = static_params
            body = body_params

        LOGGER.info('Stream: {}, Site: {}, Type: {} - Batch Sync start, Offset: {}'.format(
            stream_name,
            site,
            sub_type,
            offset))

        # Squash params to query-string params
        querystring = None
        if params.items():
            querystring = '&'.join(['%s=%s' % (key, value) for (key, value) in params.items()])
        LOGGER.info('URL for Stream: {}, Site: {} ({}): {}/{}{}'.format(
            stream_name,
            site,
            api_method,
            BASE_URL,
            path,
            '?{}'.format(querystring) if querystring else ''))
        if body and not body == {}:
            LOGGER.info('body = {}'.format(body))

        # API request data, endpoint = stream_name passed to client for metrics logging
        data = {}
        if api_method == 'GET':
            data = client.get(
                path=path,
                params=querystring,
                endpoint=stream_name)
        elif api_method == 'POST':
            data = client.post(
                path=path,
                params=querystring,
                endpoint=stream_name,
                data=json.dumps(body))

        # time_extracted: datetime when the data was extracted from the API
        time_extracted = utils.now()
        if not data or data is None or data == {}:
            LOGGER.info('xxx NO DATA xxx')

            if bookmark_field:
                write_bookmark(state, stream_name, site, sub_type, max_bookmark_value)
            return 0 # No data results

        # Transform data with transform_json from transform.py
        transformed_data = [] # initialize the record list

        # Sites endpoint returns a single record dictionary (not a list)
        if stream_name == 'sites':
            data_list = []
            data_list.append(data)
            data_dict = {}
            data_dict[data_key] = data_list
            data = data_dict
        if data_key in data:
            transformed_data = transform_json(
                data,
                stream_name,
                data_key,
                site,
                sub_type,
                dimensions_list)[data_key]
        else:
            LOGGER.info('Number of raw data records: 0')
        if not transformed_data or transformed_data is None:
            LOGGER.info('xxx NO TRANSFORMED DATA xxx')

            if bookmark_field:
                write_bookmark(state, stream_name, site, sub_type, max_bookmark_value)
            return 0 # No data results
        for record in transformed_data:
            for key in id_fields:
                if not record.get(key):
                    primary_keys_only = { id_field: record.get(id_field) for id_field in id_fields }
                    raise ValueError('Missing key {} in record with primary keys {}'.format(key, primary_keys_only))
        batch_count = len(transformed_data)

        # Process records and get the max_bookmark_value and record_count for the set of records
        max_bookmark_value = process_records(
            catalog=catalog,
            stream_name=stream_name,
            records=transformed_data,
            time_extracted=time_extracted,
            bookmark_field=bookmark_field,
            max_bookmark_value=max_bookmark_value,
            last_datetime=last_datetime)

        # to_rec: to record; ending record for the batch
        to_rec = offset + limit
        if to_rec > total_records:
            to_rec = total_records

        LOGGER.info('Stream: {}, Site: {}, Type: {}, Page: {}, Batch records: {} to {}'.format(
            stream_name,
            site,
            sub_type,
            page,
            offset,
            to_rec))
        # Pagination: increment the offset by the limit (batch-size)
        offset = offset + limit
        total_records = total_records + batch_count
        page = page + 1

    # Update the state with the max_bookmark_value for the stream, site, sub_type
    # Reference: https://developers.google.com/webmaster-tools/search-console-api-original/v3/searchanalytics/query
    # NOTE: Results are sorted by click count descending.
    #       If two rows have the same click count, they are sorted in an arbitrary way.
    #       Records are NOT sorted in DATE order.
    # THEREFOR: State is updated after ALL pages of data for stream, site, sub_type, date window
    if bookmark_field:
        write_bookmark(state,
                       stream_name,
                       site,
                       sub_type,
                       max_bookmark_value)

    # Return total_records across all batches
    return total_records


# Currently syncing sets the stream currently being delivered in the state.
# If the integration is interrupted, this state property is used to identify
#  the starting point to continue from.
# Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
def update_currently_syncing(state, stream_name):
    if (stream_name is None) and ('currently_syncing' in state):
        del state['currently_syncing']
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


def sync(client, config, catalog, state):
    start_date = config.get('start_date')

    # Get selected_streams from catalog, based on state last_stream
    #   last_stream = Previous currently synced stream, if the load was interrupted
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info('last/currently syncing stream: {}'.format(last_stream))
    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    LOGGER.info('selected_streams: {}'.format(selected_streams))

    if not selected_streams or selected_streams == []:
        return

    # Get current datetime (now_dt_str) for query parameters
    now_dttm = utils.now()
    now_dt_str = strftime(now_dttm)[0:10]
    # Reference: https://support.google.com/webmasters/answer/96568?hl=en
    # There is some delay/lag in Google Search Console results reconcilliation
    attribution_start_dttm = now_dttm - timedelta(days=ATTRIBUTION_DAYS)

    # Loop through selected_streams
    for stream_name in selected_streams:
        LOGGER.info('STARTED Syncing: {}'.format(stream_name))
        update_currently_syncing(state, stream_name)
        write_schema(catalog, stream_name)
        endpoint_config = STREAMS[stream_name]
        bookmark_field = next(iter(endpoint_config.get('replication_keys', [])), None)
        body_params = endpoint_config.get('body', {})
        endpoint_total = 0
        # Initialize body
        body = endpoint_config.get('body', {})
        # Loop through sites from config site_urls
        site_list = []
        if 'site_urls' in config:
            site_list = config['site_urls'].replace(" ", "").split(",")
        for site in site_list:
            # Skip/ignore sitemaps for domain property sites
            # Reference issue: https://github.com/googleapis/google-api-php-client/issues/1607
            #   "...sitemaps API does not support domain property urls at this time."
            if stream_name == 'sitemaps' and site[0:9] == 'sc-domain':
                LOGGER.info('Skipping Site: {}'.format(site))
                LOGGER.info('  Sitemaps API does not support domain property urls at this time.')

            else: # Not sitemaps and sites = sc-domain
                LOGGER.info('STARTED Syncing: {}, Site: {}'.format(stream_name, site))
                site_total = 0
                site_encoded = quote(site, safe='')
                path = endpoint_config.get('path').format(site_encoded)

                # Set dimension_list for performance_reports
                if stream_name == 'performance_report_custom':
                    dimensions_list = []
                    # Create dimensions_list from catalog breadcrumb
                    stream = catalog.get_stream(stream_name)
                    mdata = metadata.to_map(stream.metadata)
                    dimensions_all = ['date', 'country', 'device', 'page', 'query']
                    for dim in dimensions_all:
                        if singer.should_sync_field(singer.metadata.get(mdata, ('properties', dim), 'inclusion'),
                                                    singer.metadata.get(mdata, ('properties', dim), 'selected')):
                            # metadata is selected for the dimension
                            dimensions_list.append(dim)
                    body_params['dimensions'] = dimensions_list
                dimensions_list = body_params.get('dimensions')
                LOGGER.info('stream: {}, dimensions_list: {}'.format(stream_name, dimensions_list))

                # loop through each sub type
                sub_types = endpoint_config.get('sub_types', ['self'])
                for sub_type in sub_types:
                    sub_type_total = 0

                    # Initialize date window
                    if stream_name.startswith('performance_report'):
                        reports_dttm_str = get_bookmark(
                            state,
                            stream_name,
                            site,
                            sub_type,
                            start_date)

                        reports_dttm = strptime_to_utc(reports_dttm_str)
                        if reports_dttm < attribution_start_dttm:
                            start_dttm = reports_dttm
                        else:
                            start_dttm = attribution_start_dttm
                        end_dttm = start_dttm + timedelta(days=DATE_WINDOW_SIZE)
                        if end_dttm > now_dttm:
                            end_dttm = now_dttm

                    else:
                        start_dttm = strptime_to_utc(start_date)
                        end_dttm = now_dttm

                    # Date window loop
                    while start_dttm < now_dttm:
                        start_str = strftime(start_dttm)[0:10]
                        end_str = strftime(end_dttm)[0:10]
                        if stream_name.startswith('performance_report'):
                            body = {
                                'searchType': sub_type,
                                'startDate': start_str,
                                'endDate': end_str,
                                **body_params
                            }
                        else:
                            body = None

                        LOGGER.info('START Syncing Stream: {}, Site: {}, Type: {}, {} to {}'.format(
                            stream_name, site, sub_type, start_str, end_str))
                        total_records = sync_endpoint(
                            client=client,
                            catalog=catalog,
                            state=state,
                            start_date=start_date,
                            stream_name=stream_name,
                            site=site,
                            sub_type=sub_type,
                            dimensions_list=dimensions_list,
                            path=path,
                            endpoint_config=endpoint_config,
                            api_method=endpoint_config.get('api_method', 'GET'),
                            pagination=endpoint_config.get('pagination', 'none'),
                            static_params=endpoint_config.get('params', {}),
                            bookmark_field=bookmark_field,
                            data_key=endpoint_config.get('data_key', None),
                            body_params=body,
                            id_fields=endpoint_config.get('key_properties'))

                        # Increment totals
                        endpoint_total = endpoint_total + total_records
                        site_total = site_total + total_records
                        sub_type_total = sub_type_total + total_records

                        LOGGER.info('FINISHED Syncing Stream: {}, Site: {}, Type: {}, {} to {}'.format(
                            stream_name, site, sub_type, start_str, end_str))
                        LOGGER.info('  Records Synced for Date Window: {}'.format(total_records))

                        # Set next date window
                        start_dttm = end_dttm
                        end_dttm = start_dttm + timedelta(days=DATE_WINDOW_SIZE)
                        if end_dttm > now_dttm:
                            end_dttm = now_dttm
                        # End date window loop

                    LOGGER.info('FINISHED Syncing Stream: {}, Site: {}, Type: {}'.format(
                        stream_name, site, sub_type))
                    LOGGER.info('  Records Synced for Type: {}'.format(sub_type_total))
                    # End sub-type loop
                # End else: Not sitemaps and sites = sc-domain

                LOGGER.info('FINISHED Syncing Stream: {}, Site: {}'.format(stream_name, site))
                LOGGER.info('  Records Synced for Site: {}'.format(site_total))
                # End site loop

        LOGGER.info('FINISHED Syncing Stream: {}'.format(stream_name))
        LOGGER.info('  Records Synced for Stream: {}'.format(endpoint_total))
        update_currently_syncing(state, None)
        # End stream loop
